/*
 * MHTagger.java
 *
 * Created on 2006~52, W 2:35
 */

/**
 *
 * @author ythuang
 */

import java.util.*;
public class MHTagger {
    
    private Vector SNPs;    //the set of SNPs
    private Vector minorhap, majorhap;  //H_T and H_T bar
    private StringBuffer TSNP;  //target SNP to be replaced
    private int NumOfHap;
    
    /** Creates a new instance of MHTagger */
    public MHTagger(Vector vv, StringBuffer TSNP) {
        this.SNPs = vv;
        this.TSNP=TSNP;
        this.NumOfHap = TSNP.length();
        this.setMajorMinorHap();
        this.filter();
        //System.out.println("Target SNP: "+TSNP);
        //System.out.println("After filtering: "+SNPs);
    }
    
    //consider minor allele only now
    //identify the set of haplotypes 
    private void setMajorMinorHap(){
        //identify two set of haplotypes containing major and minor alleles observed at TSNP
        this.minorhap = new Vector();
        this.majorhap = new Vector();
        for(int i=0; i<this.TSNP.length(); i++){
            if(this.TSNP.charAt(i) == '1')
                this.minorhap.add(new Integer(i));
            else if(this.TSNP.charAt(i) == '0')
                this.majorhap.add(new Integer(i));
        }
    }
    
    private void filter(){
        //identify infeasible SNPs
        Vector temp = new Vector();
        for(int i=0; i<this.SNPs.size(); i++){
            StringBuffer sb = (StringBuffer) this.SNPs.elementAt(i);
            
            //check if there exist any mixed major and minor alleles at this SNP
            int index1 = ((Integer)this.minorhap.elementAt(0)).intValue();
            for(int j=1; j<this.minorhap.size(); j++){
                int index2 = ((Integer)this.minorhap.elementAt(j)).intValue();
                if(sb.charAt(index2) != sb.charAt(index1)){
                    //this SNP contains mixed alleles
                    temp.add(new Integer(i));
                    break;
                }
            }
        }
        
        //filter infeasible SNPs from right to left for avoiding wrong indices
        for(int i=temp.size()-1; i>=0; i--){
            SNPs.remove( ((Integer) temp.elementAt(i)).intValue());
        }
    }
      
    
    //execute the greedy algorithm to find a minimum set of SNPs which can replace TSNP
    //return indices of selected SNPs
    public Vector execute(){
        Vector Collection = this.reformuate();
        Vector solution = new Vector();
        //no SNPs can replace this one
        if(SNPs.size() == 0 || Collection.size() == 0){
            solution.add(TSNP.toString());
            return solution;
        }
        
        //initialize a boolean array representing the status of covering major hap
        boolean[] uncover = new boolean[this.NumOfHap];
        for(int i=0; i<this.majorhap.size(); i++){
            uncover[((Integer)this.majorhap.elementAt(i)).intValue()]=true;
        }
        int uncovernum = this.majorhap.size();
        
        while(uncovernum > 0){
            //find a maximum cover set
            int maxcount = 0, MaxIndex=-1;
            for(int i=0; i<Collection.size(); i++){
                Vector temp = (Vector)Collection.elementAt(i);
                
                //check the cover ability of this SNP
                int count = 0;
                for(int j=0; j<temp.size(); j++){
                    int index = ((Integer)temp.elementAt(j)).intValue();
                    if(uncover[index]){
                        count++;
                    }
                }
                
                //record current best choice
                if(count > maxcount){
                    maxcount = count;
                    MaxIndex = i;
                }
            }// end of test each SNP in the collection
            
            if(MaxIndex >=0){
                //find a contributive SNP
                //update the uncover status and remove the SNP from collection
                Vector temp = (Vector)Collection.elementAt(MaxIndex);
                for(int j=0; j<temp.size(); j++){
                    int index = ((Integer)temp.elementAt(j)).intValue();
                    if(uncover[index]){
                        uncover[index]=false;
                        uncovernum--;
                    }
                }
            }else{
                //no SNPs can replace the target SNP
                solution.clear();
                solution.add(TSNP.toString());
                return solution;
            }
            
            //Collection.remove(MaxIndex);
            solution.add((SNPs.elementAt(MaxIndex)).toString());
            //solution.add(new Integer(MaxIndex));
        }// end of while
        
        //System.out.println("Solution: "+solution);
        return solution;
    }
    
    //reformulate into a set cover problem instance
    //return a collection of sets
    private Vector reformuate(){
        Vector collection = new Vector();
        for(int i=0; i<this.SNPs.size(); i++){
            StringBuffer sb = (StringBuffer)this.SNPs.elementAt(i);
            Vector set = new Vector();
            char complement = sb.charAt( ((Integer) this.minorhap.elementAt(0) ).intValue() );
            for(int j=0; j<this.majorhap.size(); j++){
                if(sb.charAt( ((Integer) this.majorhap.elementAt(j) ).intValue() ) != complement){
                    set.add(this.majorhap.elementAt(j));
                }
            }
            if(!set.isEmpty())
                collection.add(set);
        }
        //System.out.println(collection);
        return collection;
    }
    
}
